import re
import os
import random
import string
from PIL import Image
import requests
import tempfile


class Channel:
	def __init__(self, id:str, url:str, name:str, followers:int=0):
		self.id = id
		self.url = url
		self.followers = followers
		trash = " - Topic"
		self.name = name[:-len(trash)] if name.endswith(trash) else name
		self.name = self.name[1:] if self.name.startswith("@") else self.name
	def __str__(self): return f'Channel({self.name})'
	def __repr__(self): return str(self)


class Thumbnail:
	def __init__(self, url:str):
		self.url = url
		self._raw = None

	def download(self, folder=None, quality=None):
		if not self._raw: self._parse()
		filepath = get_file_path(random_filename(), "jpg", folder)
		self._save(self._raw, filepath, quality=quality)
		return filepath

	def _save(self, entity, filepath, quality=None):
		options = {
			"max": {"quality": 100, "subsampling": 0},
			"min": {"quality": 90, "optimize": True}
		}
		save_args = options.get(quality, {"quality": 96})
		entity.save(filepath, "JPEG", **save_args)

	def _parse(self):
		img = Image.open(requests.get(self.url, stream=True).raw)
		self._raw = img.convert("RGB")

	@property
	def temp(self):
		if not self._raw: self._parse()
		new_img = self._crop()
		file = tempfile.TemporaryFile(suffix=".jpg", delete=False).name
		self._save(new_img, file)
		return file

	def _crop(self):
		width, height = self._raw.size
		min_side = min(width, height)
		left = (width - min_side) / 2
		top = (height - min_side) / 2
		right = (width + min_side) / 2
		bottom = (height + min_side) / 2
		return self._raw.crop((left, top, right, bottom))

	def __str__(self): return str(self.url)
	def __repr__(self): return 'Thumbnail()'


def random_filename(length=8):
	return ''.join(random.choice(string.ascii_lowercase) for i in range(length))

def safe_filename(s:str, max_length:int=255) -> str:
	ntfs_characters = [chr(i) for i in range(0, 31)]
	characters = [r'"',r"\#",r"\$",r"\%",r"'",r"\*",r"\,",r"\.",r"\/",r"\:",r'"',r"\;",r"\<",r"\>",r"\?",r"\\",r"\^",r"\|",r"\~",r"\\\\"]
	pattern = "|".join(ntfs_characters + characters)
	regex = re.compile(pattern, re.UNICODE)
	filename = regex.sub("", s)
	return filename[:max_length].rsplit(" ", 0)[0]

def get_file_path(filename, prefix, folder=""):
	filename = f"{safe_filename(filename)}.{prefix}"
	folder = os.path.abspath(folder) if folder else os.getcwd()
	path = os.path.join(folder, filename)
	os.makedirs(folder, exist_ok=True)
	return file_exists(path)

def file_exists(file:str) -> str:
	if os.path.exists(file):
		name, extension = os.path.splitext(file)

		match = re.search(r'\((\d+)\)$', name)
		if match:
			number = int(match.group(1)) + 1
			new_name = re.sub(r'\(\d+\)$', f'({number})', name)
		else:
			new_name = f"{name} (1)"
		return file_exists(new_name+extension)
	return file

def to_seconds(kwargs: dict) -> int:
	hour = int(kwargs.get("hour", 0))
	minute = int(kwargs.get("min", 0))
	sec = int(kwargs.get("sec", 0))
	return (hour*3600) + (minute*60) + sec

def convert_to_netscape(cookie_data):
	netscape_cookie = "# Netscape HTTP Cookie File\n\n"
	for cookie in cookie_data:
		domain = cookie.get("domain", "")
		name = cookie.get("name", "")
		value = cookie.get("value", "")
		expires = cookie.get("expirationDate", 0)

		netscape_cookie += f"{domain}\tTRUE\t/\tFALSE\t{expires}\t{name}\t{value}\n"
	return netscape_cookie

def get_cookie_file(json_cookies):
	cookies_netscape = convert_to_netscape(json_cookies)
	cookie_file = tempfile.TemporaryFile(suffix=".txt", delete=False).name
	with open(cookie_file, 'w') as f:
		f.write(cookies_netscape)
	return cookie_file



class Subtitles:
	def __init__(self, lang_code:str, lang_name:str, ext:str, url:str):
		self.code = lang_code
		self.name = lang_name
		self.extension = ext
		self.url = url

	def __str__(self): return f"Subtitles({self.code})"
	def __repr__(self): return str(self)


class SubtitlesManager:
	def __init__(self, subtitles_dict=None):
		self.subtitles = []
		self.raw_subtitles = subtitles_dict or {}
		for k, v in self.raw_subtitles.items():
			filtered = list(
				filter(
					lambda s: s.get('ext') == 'vtt' and s.get('protocol') != "m3u8_native"
				, v)
			)
			if len(filtered) > 0:
				sub = filtered[0]
				obj = Subtitles(
					lang_code=k.lower(),
					lang_name=sub.get('name'),
					ext=sub.get('ext'),
					url=sub.get('url')
				)
				self.subtitles.append(obj)

	def get(self, lang_code:str) -> Subtitles:
		'''Get subtitle by iso-code | returns Subtitles() or None'''
		code = str(lang_code.lower())
		return next((x for x in self.subtitles if x.code == code), None)


	def __str__(self): return f"Subtitles({len(self.subtitles)})"
	def __repr__(self): return str(self)
	def __len__(self): return len(self.subtitles)
	def __iter__(self): return iter(self.subtitles)



import os
import json
import subprocess
from .utils import Channel
from .utils import get_cookie_file

class CommentsManager:
	def __init__(self, link, count, cookies=None, yt_dlp="yt-dlp"):
		self.link = link
		self.count = count
		self.cookies = cookies
		self.yt_dlp = yt_dlp
		self.data = []
	def __str__(self): return f"Comments({self.count})"
	def __len__(self): return self.count
	def get(self) -> list:
		if not self.data:
			cookie_file = get_cookie_file(self.cookies) if self.cookies else None
			cmd = [
				self.yt_dlp,
				"--quiet",
				"--no-playlist",
				"--no-warnings",
				"--get-comments",
				"--skip-download",
				*(["--cookies", cookie_file] if cookie_file else []),
				"--dump-single-json",
				self.link
			]
			result = subprocess.run(cmd, capture_output=True, text=True)
			_vid_info = json.loads(result.stdout)
			self.data = _vid_info.get("comments")

			if cookie_file and os.path.exists(cookie_file): os.remove(cookie_file)
		return list(map(lambda x: Comment(x, video_url=self.link), self.data))

class Comment:
	def __init__(self, args, video_url):
		self.args = args
		self.id = args.get("id")
		self.url = f"{video_url}&lc={args.get('id')}"
		self.text = str(args.get("text", ""))
		self.likes = int(args.get("like_count", 0))
		self.is_pinned = args.get("is_pinned", False)
		self._parent = "root" if args.get("parent") == "root" else f"{video_url}&lc={args.get('parent')}"
	
	def __str__(self): return self.text
	def __repr__(self): return f"Comment({self.id})"
	@property
	def author(self) -> Channel:
		id = self.args.get("author_id")
		url = self.args.get("author_url")
		name = self.args.get("author")
		return Channel(id=id, url=url, name=name)



from .youtube import YouTube
from .utils import Thumbnail, Channel


class PlaylistVideo:
	'''Lazy load video'''
	def __init__(self, title:str, link:str, duration:int, channel:Channel, thumbnail:Thumbnail, cookies:list=None, yt_dlp="yt-dlp"):
		self.yt_obj = None
		self.cookies = cookies
		self.link = link
		self.title = title
		self.duration = duration
		self.thumbnail = thumbnail
		self.channel = channel
		self.author = channel.name
		self.yt_dlp = yt_dlp

	def load(self) -> YouTube:
		if not self.yt_obj:
			self.yt_obj = YouTube(self.link, cookies=self.cookies, yt_dlp=self.yt_dlp)
		return self.yt_obj

	def __str__(self): return f"PlaylistVideo({self.link})"
	def __repr__(self): return str(self)

class DeletedVideo:
	def __init__(self, link):
		self.link = link
	def __bool__(self): return False
	def load(self): return self


class PlaylistManager:
	def __init__(self, array=None):
		self.arr = array or []
		self.index = 0
	def __len__(self): return len(self.arr)
	def __getitem__(self, key):
		if isinstance(key, slice):
			return (self.arr[i].load() for i in range(*key.indices(len(self.arr))))
		elif isinstance(key, int):
			return self.arr[key].load()
		raise TypeError("Invalid argument type")

	def __iter__(self):
		self.index = 0
		self.end = len(self.arr)
		return self
	def __next__(self) -> YouTube:
		if self.index >= self.end: raise StopIteration
		item = self.arr[self.index]
		self.index += 1
		return item.load()



from .utils import get_file_path
from .downloader import Downloader


class Stream:
	def __init__(self, itag:str, url:str, filesize:int, metadata:dict=None):
		self.itag = itag
		self.url = url
		self.filesize = filesize
		self.metadata = metadata or {}
		self.isVideo = False
		self.isAudio = False
		self.isMuxed = False

	def get(self, attribute_name, default=0):
		return getattr(self, attribute_name, default)

	@property
	def w(self): return self.width
	@property
	def h(self): return self.height
	@property
	def res(self): return min(self.height, self.width)
	@property
	def lang(self): return self.language
	@property
	def abr(self): return self.audioBitrate
	@property
	def asr(self): return self.audioSamplerate
	
	def add_video_info(self,
		videoCodec:str,
		videoExt:str,
		width:int,
		height:int,
		fps:int
	):
		self.videoCodec = videoCodec
		self.videoExt = videoExt
		self.width = width
		self.height = height
		self.fps = fps
		self.isVideo = True

	def add_audio_info(self,
		audioCodec:str,
		audioExt:str,
		language:str,
		audioBitrate:int,
		audioSamplerate:int
	):
		self.audioCodec = audioCodec
		self.audioExt = audioExt
		self.language = language
		self.audioBitrate = audioBitrate
		self.audioSamplerate = audioSamplerate
		self.isAudio = True

	def add_muxed_info(self,
		videoCodec:str,
		videoExt:str,
		width:int,
		height:int,
		fps:int,
		audioCodec:str,
		language:str,
		audioSamplerate:int
	):
		self.videoCodec = videoCodec
		self.videoExt = videoExt
		self.width = width
		self.height = height
		self.fps = fps
		self.audioCodec = audioCodec
		self.language = language
		self.audioSamplerate = audioSamplerate
		self.isMuxed = True

	def __str__(self):
		if self.isVideo:
			return f"Video({self.width}x{self.height}.{self.videoExt} [{self.fps}fps] {self.videoCodec})"
		elif self.isAudio:
			return f"Audio({self.audioBitrate}.{self.audioExt}{' ['+self.language+']' if self.language else ''} {self.audioCodec})"
		elif self.isMuxed:
			return f"Muxed({self.width}x{self.height}.{self.videoExt} [{self.fps}fps]{' ['+self.language+']' if self.language else ''} {self.videoCodec}+{self.audioCodec})"
		else:
			return f"Stream(id={self.itag})"

	def __repr__(self):
		if self.isVideo:
			return f"Video({self.res}p)"
		elif self.isAudio:
			return f"Audio({self.audioBitrate}kbps)"
		elif self.isMuxed:
			return f"Muxed({self.res}p)"
		else:
			return f"Stream({self.itag})"


	async def download(self, *args, **kwargs) -> str:
		if self.isVideo or self.isMuxed:
			dwnl = Downloader(video=self, metadata=self.metadata)
		else:
			dwnl = Downloader(audio=self, metadata=self.metadata)
		return await dwnl(*args, **kwargs)
	


import os
import json
import subprocess
from .playlist_manager import PlaylistManager, PlaylistVideo, DeletedVideo
from .utils import Thumbnail, Channel
from .utils import get_cookie_file


class Playlist:
	def __init__(self, link, cookies:list=None, yt_dlp="yt-dlp"):
		self.link = link
		self._info = None
		self._videos = []
		self.cookies = cookies
		self.yt_dlp = yt_dlp
		cookie_file = get_cookie_file(cookies) if cookies else None
		cmd = [
			self.yt_dlp,
			"--quiet",
			"--flat-playlist",
			"--skip-download",
			"--no-warnings",
			*(["--cookies", cookie_file] if cookie_file else []),
			"--dump-single-json",
			link
		]
		result = subprocess.run(cmd, capture_output=True, text=True)
		self._info = json.loads(result.stdout)

		if cookie_file and os.path.exists(cookie_file): os.remove(cookie_file)
		if self._info.get('_type') != 'playlist': raise TypeError(f"[Not playlist]: {link}")

		for item in self._info.get('entries'):
			if not item.get('duration'):
				self._videos.append(DeletedVideo(link=item.get('url')))
			else:
				best_image = max(item.get('thumbnails'), key=lambda img: img['width']*img['height'])
				channel = Channel(
					id=item.get('channel_id'),
					url=item.get('channel_url'),
					name=item.get('channel')
				)
				self._videos.append(
					PlaylistVideo(
						title=item.get('title'),
						link=item.get('url'),
						duration=int(item.get('duration')),
						channel=channel,
						thumbnail=Thumbnail(best_image.get('url')),
						cookies=self.cookies,
						yt_dlp=self.yt_dlp
					)
				)

	@property
	def playlistId(self) -> str:
		return str(self._info.get('id'))
	@property
	def title(self) -> str:
		return str(self._info.get('title'))
	@property
	def description(self) -> str:
		return str(self._info.get('description'))
	
	@property
	def type(self):
		'''"private" or "public"'''
		return "private" if self._info.get('availability') == "private" else "public"

	@property
	def videos(self) -> PlaylistManager:
		return PlaylistManager([v for v in self._videos if isinstance(v, PlaylistVideo)])
	@property
	def raw_videos(self) -> list: return self._videos

	def __str__(self): return f"Playlist(«{self.title}»)"
	def __repr__(self): return str(self)



import os
import re
import aiohttp
import tempfile
import subprocess
from .utils import get_file_path, to_seconds


class Downloader:
	def __init__(self, video:"Stream"=None, audio:"Stream"=None, metadata:dict=None):
		self.videoStream = video if (video and (video.isVideo or video.isMuxed)) else None
		self.audioStream = audio if (audio and audio.isAudio) else None
		self.metadata = metadata or {}
		self.can_download = True
		self.CHUNK_SIZE = 10*1024*1024
		self.HEADERS = {
			"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.212 Safari/537.36",
			"Accept-Language": "en-US,en"
		}
		self.FFMPEG = "ffmpeg"
		self._DURATION_REG = re.compile(
			r"Duration: (?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})"
		)
		self._TIME_REG = re.compile(
			r"time=(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})"
		)
		self.startupinfo = None
		if os.name == 'nt':
			self.startupinfo = subprocess.STARTUPINFO()
			self.startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW

	async def _default_progress(self, current, total): return

	def abort(self):
		self.can_download = False

	def reserve_file(self, file):
		with open(file, 'w') as f:
			f.write("0")

	def remove_file(self, file):
		if os.path.exists(file): os.remove(file)

	async def __call__(self,
		output_folder:str=None,
		filename:str=None,
		video_ext:str=None,
		audio_ext:str="mp3",
		add_audio_metadata:bool=True,
		on_progress=None,
		ffmpeg_progress=None,
		on_abort=None,
		on_success=None
	) -> str:
		extension = (self.videoStream and self.videoStream.videoExt) or (self.audioStream and self.audioStream.audioExt)
		target_ext = video_ext or extension if self.videoStream else audio_ext or extension
		target_filename = filename or self.metadata.get("title", "")
		target_filepath = get_file_path(
			filename=target_filename, prefix=target_ext, folder=output_folder
		)
		self.reserve_file(target_filepath)
		on_progress = on_progress or self._default_progress
		ffmpeg_progress = ffmpeg_progress or self._default_progress

		def on_finish(f):
			if on_success: on_success(f)
			return f


		if self.videoStream and self.audioStream:
			filesize = self.videoStream.filesize + self.audioStream.filesize

			async def handle_progress(current, total, base_size=0):
				await on_progress(base_size+current, filesize)

			video_temp = tempfile.TemporaryFile(delete=False).name
			audio_temp = tempfile.TemporaryFile(delete=False).name

			await self._download_stream(self.videoStream.url, video_temp, lambda c,t: handle_progress(c,t))
			await self._download_stream(self.audioStream.url, audio_temp, lambda c,t: handle_progress(c,t,self.videoStream.filesize))
			await self._mix(video_temp, audio_temp, target_filepath, ffmpeg_progress)

			self.remove_file(video_temp)
			self.remove_file(audio_temp)


		elif self.videoStream:
			video_temp = tempfile.TemporaryFile(delete=False).name
			await self._download_stream(self.videoStream.url, video_temp, on_progress)
			if extension == target_ext:
				os.replace(video_temp, target_filepath)
			else:
				await self._convert(video_temp, target_filepath, ffmpeg_progress)

			self.remove_file(video_temp)


		elif self.audioStream:
			audio_temp = tempfile.TemporaryFile(delete=False).name
			await self._download_stream(self.audioStream.url, audio_temp, on_progress)
			if extension == target_ext and not add_audio_metadata:
				os.replace(audio_temp, target_filepath)
			else:
				await self._convert(audio_temp, target_filepath, ffmpeg_progress, (self.metadata if add_audio_metadata else None))

			self.remove_file(audio_temp)

		if self.can_download: return on_finish(target_filepath)
		self.remove_file(target_filepath)
		if on_abort: on_abort()


	async def _mix(self, video, audio, target, progress=None):
		self.remove_file(target)
		if not self.can_download: return
		codecs = ["-c:v", "copy"]
		if target.endswith(".mp4") or target.endswith(".m4a"):
			codecs.extend(["-c:a", "libmp3lame"])
		await self._ffmpeg(["-i", video, "-i", audio, *codecs, target], progress)


	async def _convert(self, inputFile, output, progress=None, metadata=None):
		self.remove_file(output)
		if not self.can_download: return
		codecs = []
		need_detele_thrumb = False
		if output.endswith(".mp3"):
			if metadata:
				if metadata.get('thumbnail'):
					thumb = metadata.get('thumbnail').temp
					need_detele_thrumb = True
					codecs.extend(["-i", thumb, "-map", "0:0", "-map", "1:0"])
				codecs.extend(["-ar", "48000", "-b:a", "192k"])
				if metadata.get('title'):
					title = metadata.get('title').replace('"', '')
					codecs.extend(["-metadata", f"title={title}"])
				if metadata.get('author'):
					artist = metadata.get('author').replace('"', '')
					codecs.extend(["-metadata", f"artist={artist}"])
				codecs.extend(["-id3v2_version", "3"])
				
		await self._ffmpeg(["-i", inputFile, *codecs, output], progress)
		if need_detele_thrumb: os.remove(thumb)


	async def _download_stream(self, url, filename, on_progress=None):
		if not self.can_download: return
		on_progress = on_progress or self._default_progress
		async with aiohttp.ClientSession(headers=self.HEADERS) as session:
			resp_head = await session.get(url)
			if not resp_head.ok:
				raise Exception(f"HTTP: {resp_head.status} {resp_head.reason}")
			
			file_size = int(resp_head.headers.get('Content-Length'))
			if not (file_size > 0):
				raise Exception("Stream filesize is 0")

			downloaded = 0
			await on_progress(downloaded, file_size)
			with open(filename, "wb") as file:
				while downloaded < file_size:
					if self.can_download:
						stop_pos = min(downloaded + self.CHUNK_SIZE, file_size) - 1
						chunk = await self._make_request(session, url+f"&range={downloaded}-{stop_pos}")
						if not chunk: break
						file.write(chunk)
						downloaded += len(chunk)
						await on_progress(downloaded, file_size)
					else: break

	async def _make_request(self, session, url, retries=3):
		for attempt in range(retries):
			try:
				resp = await session.get(url)
				chunk = await resp.content.read()
				return chunk
			except aiohttp.ClientPayloadError as e:
				if attempt == retries - 1:
					raise e


	async def _ffmpeg(self, command, on_progress=None):
		on_progress = on_progress or self._default_progress
		total_duration = 0
		if not self.can_download: return 1
		process = subprocess.Popen([self.FFMPEG, "-hide_banner"] + command, encoding='utf-8', universal_newlines=True, startupinfo=self.startupinfo, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		with process.stdout as pipe:
			history = []
			for raw_line in pipe:
				if not self.can_download:
					process.terminate()
					process.wait()
					return 1
				line = raw_line.strip()
				history.append(line)
				if total_duration == 0:
					if "Duration:" in line:
						match = self._DURATION_REG.search(line)
						total_duration = to_seconds(match.groupdict())
						await on_progress(0, total_duration)
				else:
					if "time=" in line:
						match = self._TIME_REG.search(line)
						if match:
							current = to_seconds(match.groupdict())
							await on_progress(current, total_duration)
		process.wait()
		if process.returncode != 0:
			print("\n".join(history))
			raise RuntimeError(f"FFMPEG error occurred: [{history[-1]}]")
		return process.returncode



from .stream import Stream

class StreamsManager:
	def __init__(self, streams:list=None):
		self.streams = streams if streams else []

	def __str__(self): return f"StreamsManager({len(self.streams)} streams)"
	def __repr__(self): return str(self)
	def __len__(self): return len(self.streams)
	def __iter__(self): return iter(self.streams)

	def parse(self, formats:list, metadata:dict=None) -> None:
		for format in formats:
			if format.get('format_note') == 'storyboard': continue
			if format.get('protocol') == "m3u8_native": continue

			allow_append = True
			stream = Stream(
				itag=str(format.get('format_id')),
				url=format.get('url'),
				filesize = int(format.get('filesize') or format.get('filesize_approx')),
				metadata=metadata
			)
			if format.get('acodec') != 'none' and format.get('vcodec') != 'none':
				stream.add_muxed_info(
					videoCodec=format.get('vcodec'),
					videoExt=format.get('video_ext'),
					width=int(format.get('width')),
					height=int(format.get('height')),
					fps=int(format.get('fps')),
					audioCodec=format.get('acodec'),
					language=format.get('language'),
					audioSamplerate=int(format.get('asr'))
				)
			elif format.get('vcodec') != 'none':
				stream.add_video_info(
					videoCodec=format.get('vcodec'),
					videoExt=format.get('video_ext'),
					width=int(format.get('width')),
					height=int(format.get('height')),
					fps=int(format.get('fps'))
				)
			elif format.get('acodec') != 'none':
				stream.add_audio_info(
					audioCodec=format.get('acodec'),
					audioExt=format.get('audio_ext'),
					language=format.get('language'),
					audioBitrate=int(format.get('abr')),
					audioSamplerate=int(format.get('asr'))
				)
			else:
				allow_append = False

			if allow_append:
				self.streams.append(stream)

	def filter(self,
		only_video:bool=None,
		only_audio:bool=None,
		only_muxed:bool=None,
		no_muxed:bool=None,
		max_res:int=None, # max video res
		min_res:int=None, # min video res
		max_fps:int=None, # max video fps
		min_fps:int=None, # min video fps
		custom=None # custom filter function
	) -> "StreamsManager":
		filtered = self.streams
		if only_video:
			filtered = filter(lambda x: x.isVideo or x.isMuxed, filtered)
		elif only_audio:
			filtered = filter(lambda x: x.isAudio, filtered)
		elif only_muxed:
			filtered = filter(lambda x: x.isMuxed, filtered)

		if no_muxed:
			filtered = filter(lambda x: not x.isMuxed, filtered)

		if max_res:
			filtered = filter(lambda x: x.get("res") <= max_res, filtered)
		if min_res:
			filtered = filter(lambda x: x.get("res") >= min_res, filtered)
		if max_fps:
			filtered = filter(lambda x: x.get("fps") <= max_fps, filtered)
		if min_fps:
			filtered = filter(lambda x: x.get("fps") >= min_fps, filtered)

		if custom: filtered = filter(custom, filtered)
		return StreamsManager(list(filtered))

	def order_by(self, attr:str, sub:str="", reverse=True) -> "StreamsManager":
		if sub:
			order_func = lambda x: (x.get(attr), x.get(sub))
		else:
			order_func = lambda x: x.get(attr)
		return StreamsManager(sorted(self.streams, key=order_func, reverse=reverse))

	def reverse(self) -> "StreamsManager":
		return StreamsManager(list(reversed(self.streams)))

	def first(self) -> Stream:
		if len(self) > 0:
			return self.streams[0]
		raise ValueError("No Streams")

	def last(self) -> Stream:
		if len(self) > 0:
			return self.streams[-1]
		raise ValueError("No Streams")

	def best_video(self) -> Stream:
		return self.filter(only_video=True, no_muxed=True).order_by("res", "fps").first()
	def best_audio(self) -> Stream:
		return self.filter(only_audio=True).order_by("audioBitrate").first()

	def get(self, itag:str) -> Stream:
		"""Get stream by itag"""
		stream = next((x for x in self.streams if x.itag == str(itag)), None)
		if stream: return stream
		raise ValueError(f"Stream({itag}) not found")



import setuptools
with open('README.md', 'r', encoding='utf-8') as fh:
	long_description = fh.read()

setuptools.setup(
	name='MyTube_dlp',
	version='2.0.0',
	author='Super_Zombi',
	author_email='super.zombi.yt@gmail.com',
	description='Python YouTube Downloader',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/SuperZombi/MyTube',
	project_urls={
		'Documentation': 'https://superzombi.github.io/MyTube/',
	},
	packages=['MyTube'],
	install_requires=["requests", "aiohttp", "Pillow", "yt-dlp"],
	include_package_data=True,
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	python_requires='>=3.9',
)


__version__ = "2.0.0"

from .youtube import YouTube
from .downloader import Downloader
from .stream import Stream
from .streams_manager import StreamsManager
from .subtitles import SubtitlesManager, Subtitles
from .utils import Channel, Thumbnail
from .playlist import Playlist
from .playlist_manager import PlaylistVideo, DeletedVideo
from .comments import Comment



import os
import json
import subprocess
from datetime import datetime
from .utils import Channel, Thumbnail
from .utils import get_cookie_file
from .streams_manager import StreamsManager
from .subtitles import SubtitlesManager
from .comments import CommentsManager
from .downloader import Downloader


class YouTube:
	def __init__(self, link, cookies:list=None, yt_dlp="yt-dlp"):
		self.link = link
		self._url = ""
		self._vid_info = None
		self._formats = None
		self.cookies = cookies
		self.yt_dlp = yt_dlp
		cookie_file = get_cookie_file(cookies) if cookies else None
		cmd = [
			self.yt_dlp,
			"--quiet",
			"--no-playlist",
			"--no-warnings",
			*(["--cookies", cookie_file] if cookie_file else []),
			"--dump-single-json",
			self.link
		]
		result = subprocess.run(cmd, capture_output=True, text=True)
		self._vid_info = json.loads(result.stdout)
		self._url = self._vid_info.get("webpage_url")

		if cookie_file and os.path.exists(cookie_file): os.remove(cookie_file)

	def __str__(self): return f'MyTube({self.videoId})'
	def __repr__(self): return str(self)

	@property
	def videoId(self) -> str:
		return str(self._vid_info.get("id"))
	
	@property
	def title(self) -> str:
		return str(self._vid_info.get("title"))

	@property
	def author(self) -> str:
		return str(self.channel.name)

	@property
	def type(self) -> str:
		'''"video" or "music"'''
		if any(e.lower() == "music" for e in self._vid_info.get("categories")):
			return "music"
		if "music.youtube" in self.link:
			return "music"
		return "video"

	@property
	def description(self) -> str:
		return str(self._vid_info.get("description"))

	@property
	def duration(self) -> int:
		"""Duration in seconds"""
		return int(self._vid_info.get("duration"))

	@property
	def views(self) -> int:
		"""Views count"""
		return int(self._vid_info.get("view_count"))
	
	@property
	def likes(self) -> int:
		"""Likes count"""
		return int(self._vid_info.get("like_count"))
	
	@property
	def comments(self) -> CommentsManager:
		count = int(self._vid_info.get("comment_count"))
		return CommentsManager(self._url, count, cookies=self.cookies, yt_dlp=self.yt_dlp)
	
	@property
	def thumbnail(self) -> Thumbnail:
		return Thumbnail(self._vid_info.get("thumbnail"))

	@property
	def upload_date(self) -> datetime:
		ts = int(self._vid_info.get("timestamp"))
		return datetime.utcfromtimestamp(ts)

	@property
	def subtitles(self) -> SubtitlesManager:
		return SubtitlesManager(self._vid_info.get("subtitles"))

	@property
	def streams(self) -> StreamsManager:
		self._formats = self._vid_info.get('formats', [])
		streamsManager = StreamsManager()
		streamsManager.parse(self._formats, metadata=self.metadata)
		return streamsManager

	@property
	def metadata(self) -> dict:
		return {
			"title": self.title,
			"author": self.author,
			"thumbnail": self.thumbnail
		}

	@property
	def channel(self) -> Channel:
		id = self._vid_info.get("channel_id")
		url = self._vid_info.get("channel_url")
		name = self._vid_info.get("channel")
		followers = int(self._vid_info.get("channel_follower_count"))
		return Channel(id=id, url=url, name=name, followers=followers)

	def download(self, video=None, audio=None, metadata=None) -> Downloader:
		return Downloader(video, audio, (metadata or self.metadata))
